from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import os
import sys
import collections
import numpy as np
from random import randint 

import pytest





from mmgroup.dev.hadamard.hadamard_t import bitparity
from mmgroup.dev.mm_op.mm_op import MM_Op, INT_BITS

from mmgroup.tests.test_hadamard.matrices import hadamard_matrix
from mmgroup.tests.test_hadamard.matrices import msym16
from mmgroup.tests.test_hadamard.matrices import mat_l_16
from mmgroup.tests.test_hadamard.cythonize import SIZES 
from mmgroup.tests.test_hadamard.cythonize import PRIMES 
from mmgroup.tests.test_hadamard.cythonize import build, kill
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.groups.mgroup_n import MGroupNWord


NP_DTYPE = np.uint32 if INT_BITS == 32 else np.uint64


V = SparseMmV
G = MGroupNWord


################################################################
# Build external module with code to be tests
################################################################

test_hadamard = None

def build_code():
    """Build Cython code, import is as module test_hadamard

    This function builds the Cython module
    mmgroup.tests.test_hadamard.temp.test_hadamard
    an imports that module as test_hadamard.

    Since that build process takes some time, we do it on
    demand only.
    """
    global test_hadamard 
    if test_hadamard is None:
        build()
        print("importing  test_hadamard...")
        from mmgroup.tests.test_hadamard.temp import test_hadamard as t
        print("import done")
        test_hadamard = t

################################################################
# Auxiliary functions
################################################################

LOG_BIN = {4:2, 16:4, 64:6}

def pwr2modp(k, p):
    """Return 2**k mod p for any integer k"""
    if k < 0:
        assert p & 1
        return pow((p + 1) >> 1, -k, p)
    return pow(2, k, p)

################################################################
# Wrapper for functions in library test_hadamard.pyx
################################################################

def matrix_function(name, p, *data, **kwds):
    def pack_input_vector(data, dest, offset):
        for i in range(VECTOR_INTS):
            value = 0
            part = data[i * INT_FIELDS : (i+1) * INT_FIELDS]
            for j, x in enumerate(part):
                value |= (int(x) % p) << (j * FIELD_BITS)
            dest[offset + i] = value

    try:
        verbose = int(kwds["verbose"])
    except:
        verbose = 0
    basics = MM_Op(p)
    vlen = len(data[0])
    if name[-1:].isdigit():
        function_name = "mod%d_%s_test" % (p, name)
    else:
        function_name = "mod%d_%s%d_test" % (p, name, vlen)
    f =  test_hadamard.__dict__[function_name]
    FIELD_BITS = basics.FIELD_BITS
    INT_FIELDS = basics.INT_FIELDS
    VECTOR_INTS = max(1, vlen // INT_FIELDS)
    w_len = max(10, len(data)) * VECTOR_INTS
    w = np.zeros(w_len, dtype = NP_DTYPE)
    for i, d in enumerate(data):
        pack_input_vector(d, w, i * VECTOR_INTS)
    if verbose > 1:
        print("Input to function %s, p = %d, len = %d:" % (
           name, p, vlen))
        print([hex(x) for x in w[:VECTOR_INTS]])
    f(w)
    if verbose > 1:
        print("Output of function %s:" % name)
        print([hex(x) for x in w[:VECTOR_INTS]])
    result = np.zeros(vlen, dtype = np.int32)
    for i in range(vlen):
        index, sh = divmod(i, INT_FIELDS)
        o =  int(w[index]) >> (sh * FIELD_BITS)
        result[i] = ((o & p) % p)
    return result


################################################################
# Testing multiplication with Hadamard matrix
################################################################



def hadamard_function(p, vector, verbose = 0):
    return matrix_function("hadamard", p, vector, verbose=verbose)


def one_test_hadamard_function(p, vector, verbose = 0):
    if verbose:
        print("\nTest multiplication v * H with Hadamard matrix H, v =")
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = hadamard_function(p, vector, verbose = verbose)
    k = LOG_BIN[len(vector)]
    h_matrix = hadamard_matrix(k) * pwr2modp(-k >> 1, p) % p
    expected = (vector @ h_matrix) % p
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            print("Multiplication v * H with Hadmard matrix H, v =")
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")


def hadamard_function_testcases(n_cases = 5):
    test_cases = [
        (3,  [1,0,0,0]),
        (3,  [0,1,0,0]),
        (3,  [0,0,1,0]),
        (3,  [0,0,0,1]),
    ]
    for t in test_cases:
        yield t
    for i in range(n_cases):
        for p in PRIMES:
            for l in SIZES:
                v = [randint(0, p - 1) for l in range(1 << l)]
                yield p, v 

@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_hadamard_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test C code for multiplication with Hadamard matrix")
    for p, v in hadamard_function_testcases(n_cases):
        one_test_hadamard_function(p, v, verbose)
    print("Test passed")



################################################################
# Testing exchanging entries with odd parity
################################################################


def xch_parity_function(p, vector, verbose = 0):
    return matrix_function("xch_parity", p, vector, verbose=verbose)

def ref_xch_parity(vector):
    l = len(vector)
    res = np.zeros(l, dtype = np.int32)
    for i in range(l):
        j = l - i - 1 if bitparity(i) else i
        res[i] = vector[j]
    return res

def one_test_xch_parity_function(p, vector, verbose = 0):
    if verbose:
        print("\nTest exchanging entries of v with odd parity, v =")
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = xch_parity_function(p, vector, verbose = verbose)
    k = LOG_BIN[len(vector)]
    expected = ref_xch_parity(vector) % p
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            s = "Exchanging entries of v (mod %d) with odd parity, v ="
            print(s % p)
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")

def xch_parity_function_testcases(n_cases = 5):
    test_p = [3, 7, 127]
    for p  in test_p:
        if not p in PRIMES:
            continue
        for i in range(64):
            lst = [0] * 64
            lst[i] = 1
            yield p, lst
    for i in range(n_cases):
        for p in PRIMES:
            v = [randint(0, p - 1) for l in range(64)]
            yield p, v 

@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_xch_parity_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test C code for exchanging entries with odd parity")
    for p, v in xch_parity_function_testcases(n_cases):
        one_test_xch_parity_function(p, v, verbose)
    print("Test passed")



################################################################
# Testing operation t done by 64 times 64 matrices
################################################################



def op_t64_function(p, vector, exp, verbose = 0):
    exp = exp % 3
    if exp == 0: return vector
    x = [exp - 1]
    return matrix_function("op_t", p, vector, x, verbose=verbose)


    

def ref_op_t64(p, vector, exp):
    space = V(p)
    v = space()
    for i in range(64):
        v += int(vector[i]) * space('T',0, i)
    v *= G('t', exp)
    res = np.zeros(64, dtype = np.int32)
    for i in range(64):
        res[i] = v['T', 0, i]
    return res

def one_test_op_t64_function(p, vector, exp, verbose = 0):
    if verbose:
        print("\nTest op t64, exp = ", exp, ",  v =\n")
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = op_t64_function(p, vector, exp, verbose = verbose)
    k = LOG_BIN[len(vector)]
    expected = ref_op_t64(p, vector, exp) 
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            s = "\nTest op t64 mod %d, exp = %d,  v ="
            print(s % (p, exp))
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")



def op_p64_function_testcases(n_cases = 5):
    for p  in PRIMES:
        for i in range(64):
            lst = [0] * 64
            lst[i] = 1
            yield p, lst, 1
            yield p, lst, 2
    for i in range(n_cases):
        for p in PRIMES:
            v = [randint(0, p - 1) for l in range(64)]
            yield p, v, 1 
            yield p, v, 2 


@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_op_p64_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test operator t on tag 'T' done by 64 x 64 matrices")
    for p, v, exp1 in op_p64_function_testcases(n_cases):
        one_test_op_t64_function(p, v, exp1, verbose)
    print("Test passed")




################################################################
# Testing operation t done by 3 times 3 matrices
################################################################


def v24_ints(p):
    return MM_Op(p).V24_INTS

def int_fields(p):
    return MM_Op(p).INT_FIELDS


def op_t3_function(p, vector, exp, verbose = 0):
    if exp == 0: return vector
    assert len(vector) == 3 * int_fields(p)
    exp = exp % 3
    x = [exp - 1]
    return matrix_function("op_t3", p, vector, x, verbose=verbose)


def pr_a3(p, a):
    ld = len(str(p))
    fields =  int_fields(p)
    for i in range(0, 3*fields, fields):
        for j in range(fields):
            print("%*d" % (ld,a[i+j]), end = " ")
        print("")





def ref_op_t3(p, vector, exp):
    fields = int_fields(p)
    assert len(vector) == 3 * fields
    exp = exp % 3
    space = V(p)
    result = [None] * (3 * fields)
    for i in range(fields):
        v = space()
        for tag, j in [("A", 0), ("B", fields), ("C", 2 * fields)]:
            v += int(vector[i + j]) * space(tag, 1, 0)
        v *= G('t', exp)
        for tag, j in [("A", 0), ("B", fields), ("C", 2 * fields)]:
            result[i + j] = v[tag, 1, 0]
    return result



def one_test_op_t3_function(p, vector, exp, verbose = 0):
    if verbose:
        print("\nTest op t3, exp = ", exp, ",  v =\n")
        pr_a3(p, vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = op_t3_function(p, vector, exp, verbose = verbose)
    expected = ref_op_t3(p, vector, exp) 
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            s = "\nTest op t3 mod %d, exp = %d,  v ="
            print(s % (p, exp))
            pr_a3(p,vector)
        print("Result obtained:")
        pr_a3(p, obtained)
        if not ok:
            print("Result expected:")
            pr_a3(p, expected)
            raise ValueError("Wrong result")


def op_p3_function_testcases(n_cases = 5):
    for p  in PRIMES:
        fields = int_fields(p)
        for i in range(3 * fields):
            lst = [0] * (3 * fields)
            lst[i] = 1
            yield p, lst, 1
            yield p, lst, 2
    for i in range(n_cases):
        for p in PRIMES:
            fields = int_fields(p)
            v = [randint(0, p - 1) for l in range(3 * fields)]
            yield p, v, 1 
            yield p, v, 2 


@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_op_p3_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test operator t on tags 'ABC' done by 3 x 3 matrices")
    for p, v, exp1 in op_p3_function_testcases(n_cases):
        one_test_op_t3_function(p, v, exp1, verbose)
    print("Test passed")






################################################################
# test symmetric operation xi done by 64 times 64 matrices
################################################################

msym4a = np.array(
    [[1,1,1,1], [1,1,-1,-1], [1,-1,1,-1], [1,-1,-1,1]]
)

def op_xi_sym64_function(p, vector, verbose = 0):
    assert vector.shape == (16, 24)
    vector %= p    
    v1 = np.zeros( (16, 32), dtype = np.int32)
    v1[0:16, 0:24] = vector
    v1 = v1.reshape(16*32)
    v0 = np.zeros(8, dtype = np.int8)
    v2 = matrix_function("op_xi64", p, v1, [2], verbose=verbose)
    v2 = v2.reshape( (16, 32) )
    assert np.count_nonzero(v2[: , 24:]) == 0
    return v2[: , :24] % p


def ref_op_xi_sym64(p, vector):
    v = np.copy(vector)
    for i in range(16):
        for j in range(0, 24, 4):
            v[i, j:j+4] = v[i, j:j+4].dot(msym4a)
    for i in range(24):
        v[:,i] = mat_l_16.dot(v[:,i])
    q = pow((p + 1) >> 1, 3, p)
    return v * q % p
    


def one_test_op_xi_sym64_function(p, vector, verbose = 0):
    if verbose:
        print("\nTest op lsym64 mod %d,  v =\n" % p)
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = op_xi_sym64_function(p, vector, verbose = verbose)
    expected = ref_op_xi_sym64(p, vector) 
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            print("\nTest op xi sym64 mod %d,  v =\n" % p)
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")


def op_xi_sym64_function_testcases(n_cases = 5):
    for p  in PRIMES:
        for i in range(16):
            for j in range(24):
                v = np.zeros((16,24), dtype = np.int32)
                v[i, j] = 1
                yield p, v
    for i in range(n_cases):
        for p in PRIMES:
            v = np.zeros((16,24), dtype = np.int32)
            for i in range(16):
                for j in range(24):   
                    v[i,j] = randint(0, p-1)
            yield p, v

@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_op_xi_sym64_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test part of operator xi done by 64 x 64 matrices")
    for p, v in op_xi_sym64_function_testcases(n_cases):
        one_test_op_xi_sym64_function(p, v, verbose)
    print("Test passed")

################################################################
# test operation xi done by 64 times 64 matrices
################################################################


def op_xi64_function(p, vector, exp, verbose = 0):
    exp = exp % 3
    if exp == 0: return vector
    x = [exp - 1]
    assert vector.shape == (16, 24)
    vector %= p    
    v1 = np.zeros( (16, 32), dtype = np.int32)
    v1[0:16, 0:24] = vector
    v1 = v1.reshape(16 * 32)
    v2 = matrix_function("op_xi64", p, v1, x, verbose=verbose)
    v2 = v2.reshape( (16, 32) )
    assert np.count_nonzero(v2[: , 24:]) == 0
    return v2[: , :24] % p


def ref_op_xi64(p, vector, exp):
    exp = exp % 3
    if exp == 0: return vector
    space = V(p)
    v = space(0)
    for i in range(16):
        for j in range(24):
            v += int(vector[i,j]) * space('Y',i + 0x400, j)
    v *= G('l', exp)
    #print(v)
    v1 = np.zeros( (16, 24), dtype = np.int32)
    for i in range(16):
        for j in range(24):
            v1[i, j] = v['Y', i + 0x400, j]
    return v1
    

def one_test_op_xi64_function(p, vector, exp, verbose = 0):
    if verbose:
        print("\nTest op xi 64 mod %d, exp = %d,  v =\n" % (p, exp))
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = op_xi64_function(p, vector, exp, verbose = verbose)
    expected = ref_op_xi64(p, vector, exp) 
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            print("\nTest op xi 64 mod %d, exp = %d, v =\n" % (p, exp))
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")


def op_xi64_function_testcases(n_cases = 5):
    if n_cases >= 100:
        for p  in PRIMES:
            for i in range(16):
                for j in range(24):
                    v = np.zeros((16,24), dtype = np.int32)
                    v[i, j] = 1
                    yield p, v, 1
                    yield p, v, 2
    for i in range(n_cases):
        for p in PRIMES:
            v = np.zeros((16,24), dtype = np.int32)
            for i in range(16):
                for j in range(24):   
                    v[i,j] = randint(0, p-1)
            yield p, v, 1
            yield p, v, 2

@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_op_xi64_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test operator xi on tags 'YZ' done by 64 x 64 matrices")
    for p, v, exp in op_xi64_function_testcases(n_cases):
        one_test_op_xi64_function(p, v, exp, verbose)
    print("Test passed")






################################################################
# test operation xi done by 16 times 16 matrices
################################################################

def op_xi16_function(p, vector, exp, verbose = 0):
    exp = exp % 3
    if exp == 0: return vector
    x = [exp - 1]
    assert vector.shape == (4, 24)
    vector %= p    
    v1 = np.zeros( (4, 32), dtype = np.int32)
    v1[0:16, 0:24] = vector
    v1 = v1.reshape(4 * 32)
    v2 = matrix_function("op_xi16", p, v1, x, verbose=verbose)
    v2 = v2.reshape( (4, 32) )
    assert np.count_nonzero(v2[: , 24:]) == 0
    return v2[: , :24] % p



def ref_op_xi16(p, vector, exp):
    """Generate reference vector for 16 1 14 operation on tag 'A'"""
    assert (vector[:4,:4] == vector[:4,:4].T).all()
    # The previous assertion is a rather subtile symmetry condition
    # required for the python reference implementation to work.
    # Function symmetrize_test_matrix(vector) forces that symmetry.
    exp = exp % 3
    if exp == 0: return vector
    space = V(p)
    v = space()
    for i in range(4):
        for j in range(24):
            i0, j0 = max(i,j), min(i,j)
            if j >= 4 or i >= j:
                v += int(vector[i,j]) * space('A', i, j)
    v *= G('l', exp)
    #print(v)
    v1 = np.zeros( (4, 24), dtype = np.int32)
    for i in range(4):
        for j in range(24):
            i0, j0 = max(i,j), min(i,j)
            v1[i, j] = v['A', i0, j0]
    return v1


def one_test_op_xi16_function(p, vector, exp, verbose = 0):
    if verbose:
        print("\nTest op xi16 mod %d, exp = %d,  v =\n" % (p, exp))
        print(vector)
    vector = np.array(vector, dtype = np.int32) % p
    obtained = op_xi16_function(p, vector, exp, verbose = verbose)
    expected = ref_op_xi16(p, vector, exp) 
    ok = (obtained == expected).all()
    if verbose or not ok:
        if not ok:
            print("\nTest op xi 16 mod %d, exp = %d, v =\n" % (p, exp))
            print(vector)
        print("Result obtained:")
        print(obtained)
        if not ok:
            print("Result expected:")
            print(expected)
            raise ValueError("Wrong result")


def symmetrize_test_matrix(vector, i = 0):
   """Force symmetry of 'vector' required for function ref_op_xi16()"""
   i = i & ~ 3
   for j in range(4):
       for k in range(j):
           vector[i + j, k] = vector[i + k, j]


def op_xi16_function_testcases(n_cases = 5):
    if n_cases >= 0:
        for p in [7] + PRIMES:
            for i in range(4):
                for j in range(24):
                    v = np.zeros((4,24), dtype = np.int32)
                    v[i, j] = 1
                    symmetrize_test_matrix(v)
                    yield p, v, 1
                    yield p, v, 2
    for i in range(n_cases):
        for p in PRIMES:
            v = np.zeros((4,24), dtype = np.int32)
            for i in range(4):
                for j in range(24):   
                    v[i,j] = randint(0, p-1)
            symmetrize_test_matrix(v)
            yield p, v, 1
            yield p, v, 2


@pytest.mark.compiler
@pytest.mark.slow
@pytest.mark.hadamard
def test_op_xi16_function(n_cases = 5, verbose = 0):
    build_code()
    print("Test operator xi on tag 'A' done by 16 x 16 matrices")
    for p, v, exp in op_xi16_function_testcases(n_cases):
        one_test_op_xi16_function(p, v, exp, verbose)
    print("Test passed")



################################################################
# Clean up
################################################################


################################################################
# Main test program
################################################################



N_CASES = 20
if __name__ == "__main__":
    test_hadamard_function(N_CASES, verbose = 0)  
    test_xch_parity_function(N_CASES, verbose = 0)    
    test_op_p3_function(N_CASES, verbose = 0)
    test_op_p64_function(N_CASES, verbose = 0)
    test_op_xi16_function(N_CASES, verbose = 0)
    test_op_xi_sym64_function(N_CASES, verbose = 0)
    test_op_xi64_function(N_CASES, verbose = 0)



"""Generate tables for operation xi on the rep 196884x of the monster

Operators xi and xi**2 (with xi**3 == 1) operate on unit vectors with 
tags B, C, T, X as defined in mm_aux.py. We divide the 759 octads 
occuring as indices for tag T into three groups:

 TG:  octad   0 ...  15,   15 grey even octads
 T0:  octad  15 ... 374,  360 even octads which are not grey
 T1:  octad 375 ..  758,  384 odd octads

Octads are Golay code elements. Odd, even and grey Golay code elements 
are defined in [Seys19]. In the numbering of our octads and basis
vectorv, octads are ordered as in the table above. Similarly, we divide 
the basis vectors with tag T into two blocks with tags T0 and T1. 
Basis vectors in T0 are indexed by even and are indexed by odd Golay 
code vectors.

The unit vectors for each tag have a natural matrix structure.
Let BC the union of the basis vectors with tags B, C and TG Then
Operator xi permutes tags BC, T0, T1, X0 and X1 as described in the
following table. 


Tag     Matrix      Blocks     Block    Total     Common   xi(tag)
        structure              size     size     tag name
 B      24  x  24   \
 C      24  x  24    >   1      2496(*)  2496       BC       BC
 TG     15  x  64   /
 T0    360  x  64       45    8 x 64    23040       T0       T0
 T1    384  x  64       64    6 x 64    24576       T1       X0
 X0   1024  x  24       64   16 x 24    24576       X0       X1
 X1   1024  x  24       64   16 x 24    24576       X1       T1

(*) The block sizes for tags B and C have been computed under the
    assumption that 32 contiguous locations are reserved for 
    the 24 entries of a row, as described below.

The entries corresponding to a tag are stored in row-major order,
so that e.g. T0[i,j] and T0[i,j+1] are contiguous. If a tag has
stucture I x 24 then 32 instead of 24 entries are reserved for 
each row. The order of tags is as in the table. 

The entries corresponding to tags T0, T1, X0, X1 can be divided
into 45 or 64 contiguous blocks as indicated in the table. Then 
the image of the i-th block of a tag I under the operator xi is 
the i-th block of tag xi(I). We will use this local structure of
operator xi in our implementation of operators xi and xi**2.

Tables for the monomial operation of xi can be generated by
class mat24fast.Mat24Xi. Class Pre_MM_TablesXi provides
updated version of these tables suitable for the implementation
of operator xi. Class MM_TablesXi puts these tables into an
order suitable for that implementation.

TODO: continue documentation!!!
"""

#234567890123456789012345678901234567890123456789012345678901234567890

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import os
import sys
import numpy as np
from itertools import product

from mmgroup.generate_c import UserDirective

try:
    # Try importing the fast C function
    from mmgroup import generators as gen 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.generators.gen_xi_ref import GenXi 
    gen = GenXi




def make_table_bc_symmetric(table):
    def b(i, j):
        return 32 * i + j
    def c(i, j):
        return 32 * (i + 24) + j
    for i in range(24):
        for j in range(i):
            table[b(j,i)] = table[b(i,j)] 
            table[c(j,i)] = table[c(i,j)] 
        table[b(i,i)] = b(i,i)
        table[c(i,i)] = c(i,i)


def cut24(table):
    table = table.reshape(-1, 32)
    #assert sum(table[:,24:].reshape(-1)) == 0, (table.shape, table[:,24:])
    table = table[:,:24]
    table = table.reshape(-1)
    return table



def check_table(table, blocks, row_length):
    assert row_length in (24, 32)
    if row_length == 24:
        table = cut24(np.copy(table))
    length = len(table)
    image_length = (max(table & 0x7fff) + 31) & -32
    blocklen, r0 = divmod(length, blocks)
    image_blocklen, r1 = divmod(image_length, blocks)
    assert r0 == r1 == 0
    for i, start in enumerate(range(0, length, blocklen)):
         part = table[start:start+blocklen] & 0x7fff
         t_min, t_max = i*image_blocklen, (i+1)*image_blocklen
         assert t_min <= min(part) <= max(part) < t_max



class Pre_MM_TablesXi: 
    def __init__(self):   
        BCT, T0 = 24*32, 72*32 + 15*64
        T1, X0  = 72*32 + 375*64,  72*32 + 759*64
        X1 = X0 + 1024*32

        self.TAG_NAMES = {BCT:"BC", T0:"T0", T1:"T1", X0:"X0", X1:"X1"} 

        self.SHAPES = {
            BCT: (1, 78, 32), T0: (45, 16, 32), T1: (64, 12, 32),
            X0: (64, 16, 24), X1: (64, 16, 24)
        }

        self.MAP_XI = {
            (1,BCT): BCT, (1,T0): T0, (1, T1): X0, (1, X0): X1, (1,X1): T1
        }
        MAP_XI_ = {}
        for (exp, src), dest in self.MAP_XI.items():
            MAP_XI_[2, dest] = src
        self.MAP_XI.update(MAP_XI_)
        del MAP_XI_ 

        TABLE_ORDER = [None, BCT, T0, T1, X0, X1]

        self.TABLE_INDICES = {}
        for i, ofs in enumerate(TABLE_ORDER):
            if i:
                self.TABLE_INDICES[1, ofs] = (1, i) 
                self.TABLE_INDICES[2, ofs] = (2, i) 

        self.REVERSE_TABLE_INDICES = {}
        for x, y in self.TABLE_INDICES.items():
            self.REVERSE_TABLE_INDICES[y] = x

        self.PERM_TABLES = {}
        self.SIGN_TABLES = {}
        for (i, j), (_, start) in self.REVERSE_TABLE_INDICES.items():
            image_start = self.MAP_XI[i, start]
            shape =  self.SHAPES[start]
            image_shape = self.SHAPES[image_start]
            table = gen.make_table(j, i)
            assert shape[0] == image_shape[0], (shape[0], image_shape[0]) 
            assert len(table) == shape[0] * shape[1] * 32
            image_len = image_shape[0] * image_shape[1] * 32
            check_table(table, shape[0], shape[2])
            inv_table = gen.invert_table(table, shape[2], image_len)
            if start == BCT:
                make_table_bc_symmetric(inv_table)
            t_perm, t_sign = gen.split_table(inv_table, shape[1]*32)
            del inv_table
            if image_shape[2] == 24:
                t_perm = cut24(t_perm)
            self.PERM_TABLES[i, j] = t_perm 
            self.SIGN_TABLES[i, j] = t_sign

def map_table(exp, j):
    j1 = j if (exp == 1 or j < 4) else 9 - j
    return exp, j1

def operator_(x):
    if x == 0: return None
    if x > 0: return '+'
    return '-'

#234567890123456789012345678901234567890123456789012345678901234567890


def code_pointers(name_perm, name_sign):
    s = ""
    for exp in range(1,3):
       s += "{\n"
       for i in range(1,6):
           t_p = name_perm + str(exp) + str(i) 
           t_s = name_sign + str(exp) + str(i)
           s += "    {%s, %s}%s\n" % (t_p, t_s, ("," if i < 5 else ""))
       s += "}%s\n" % ("," if exp < 2 else "")
    return s


class MM_TablesXi:
    done_ = False
    directives = {
        "CODE_XI_POINTERS": UserDirective(code_pointers, "ss"),
    }
    def __init__(self):
        cls = self.__class__
        if cls.done_:
            return
        Pre_Tables = Pre_MM_TablesXi()
        cls.TAG_NAMES = Pre_Tables.TAG_NAMES
        cls.PERM_TABLES = {}
        cls.SIGN_TABLES = {}
            
        for i in range(1,3):
            for j in range(1,6):
                i1, j1 =  map_table(i, j)
                cls.PERM_TABLES[i, j] = Pre_Tables.PERM_TABLES[i1, j1] 
                cls.SIGN_TABLES[i, j] = Pre_Tables.SIGN_TABLES[i1, j1]

        cls.SOURCE_SHAPES = {}
        cls.DEST_SHAPES = {}
        cls.SOURCE_START_1 = {}
        cls.SOURCE_START_DIFF = {}
        cls.SOURCE_OP_DIFF = {}
        cls.DEST_START_1 = {}
        cls.DEST_START_DIFF = {}
        cls.DEST_OP_DIFF = {}
        cls.SOURCE_TAGS = {}
        cls.DEST_TAGS = {}
        cls.MAX_ABS_START_DIFF = MDIFF = 32768
        REV = Pre_Tables.REVERSE_TABLE_INDICES
        for j in range(1,6):
            source_start1 = REV[map_table(1,j)][1] 
            source_start2 = REV[map_table(2,j)][1]
            dest_start1 = Pre_Tables.MAP_XI[1, source_start1]
            dest_start2 = Pre_Tables.MAP_XI[2, source_start2]
            SHAPES = Pre_Tables.SHAPES
            assert SHAPES[source_start1] == SHAPES[source_start2] 
            assert SHAPES[dest_start1] == SHAPES[dest_start2]
            cls.SOURCE_SHAPES[j] = SHAPES[source_start1]
            cls.DEST_SHAPES[j] = SHAPES[dest_start1]
            cls.SOURCE_START_1[j] = source_start1
            cls.SOURCE_START_DIFF[j] = source_start2 - source_start1
            cls.DEST_START_1[j] = dest_start1
            cls.DEST_START_DIFF[j] = dest_start2 - dest_start1
            assert abs(cls.SOURCE_START_DIFF[j]) in [0, MDIFF]
            assert abs(cls.DEST_START_DIFF[j]) in [0, MDIFF]
            cls.SOURCE_OP_DIFF[j] = operator_(cls.SOURCE_START_DIFF[j])
            cls.DEST_OP_DIFF[j] = operator_(cls.DEST_START_DIFF[j])
            cls.SOURCE_TAGS[j] = (cls.TAG_NAMES[source_start1], 
                cls.TAG_NAMES[source_start2]) 
            cls.DEST_TAGS[j] = (cls.TAG_NAMES[dest_start1], 
                cls.TAG_NAMES[dest_start2]) 
        del SHAPES; del REV
       
        TABLES = {}
        for k in product([1,2], range(1,6)):
            TABLES["MM_TABLE_PERM_XI_%d%d" % k] = cls.PERM_TABLES[k]
            TABLES["MM_TABLE_SIGN_XI_%d%d" % k] = cls.SIGN_TABLES[k]

        cls.tables = TABLES
        cls.done_ = True

    @classmethod
    def display_config(cls):
        cls.__init__()
        print("source:")
        print(" shapes:", cls.SOURCE_SHAPES)
        print(" start: ", cls.SOURCE_START_1)
        print(" diff:  ", cls.SOURCE_OP_DIFF)
        print(" tags:  ", cls.SOURCE_TAGS)
        print("destination:")
        print(" shapes:", cls.DEST_SHAPES)
        print(" start: ", cls.DEST_START_1)
        print(" diff:  ", cls.DEST_OP_DIFF)
        print(" tags:  ", cls.DEST_TAGS)



class Mockup_MM_TablesXi:
    directives = {
        "CODE_XI_POINTERS": UserDirective(code_pointers, "ss"),
    }
    TABLES = {}
    for k in product([1,2], range(1,6)):
        TABLES["MM_TABLE_PERM_XI_%d%d" % k] = [0]
        TABLES["MM_TABLE_SIGN_XI_%d%d" % k] = [0]
    tables = TABLES



if __name__ == "__main__":
     MM_TablesXi().display_config() 






"""Configutation for development"""

import sys
import os
from collections import defaultdict

###############################################################################
# Set NATIVE = True for a version optimized for your PC
# Set NATIVE = False for a generic version
NATIVE = False

###############################################################################
# Bit length of an integer used in the representation of the monster 
# This should be 32 for a 32-bit and 64 for a 64-bit system.
INT_BITS = 64

###############################################################################
# Representations of the monster are built modulo the following numbers
PRIMES = [3, 7, 15, 31, 127, 255]



###############################################################################
# Extra compile arguments for .c files to be used in setup.py
EXTRA_COMPILE_ARGS = defaultdict(list)

# Update the EXTRA_COMPILE_ARGS dictionary with entries
#     compiler : <list of options> .
# Here 'compiler' is a string. For a list of available compilers, run:
#     python setup.py build_ext --help-compiler

mingw32_compile_args =  ["-m64", "-Ofast", "-flto", "-funroll-loops"]
unix_compile_args = ["-static-libgcc",  
             "-Ofast", "-flto", "-funroll-loops"
            ]
msvc_compile_args = ["/O2"]

if NATIVE:
     mingw32_compile_args.append("-march=native") 
     unix_compile_args.append("-march=native") 

EXTRA_COMPILE_ARGS.update({
    'mingw32' : mingw32_compile_args,
    'unix' : unix_compile_args,
    'msvc': msvc_compile_args,
})


###############################################################################
# Extra link arguments for .c files to be used in setup.py
EXTRA_LINK_ARGS = defaultdict(list)
# Similiar to updating EXTRA_COMPILE_ARGS, we may update 
# the EXTRA_LINK_ARGS dictionary with a list of pairs
#     compiler : <list of options> .


unix_link_args = ["-static-libgcc"]

EXTRA_LINK_ARGS.update({
    'unix' : unix_link_args,
})

###############################################################################
# Directories

ROOT_DIR = "."
SRC_DIR = os.path.join(ROOT_DIR, "src")
PACKAGE_DIR = os.path.join(SRC_DIR, "mmgroup")
DEV_DIR = os.path.join(PACKAGE_DIR, "dev")

REAL_SRC_DIR = os.path.realpath(SRC_DIR)


C_DIR = os.path.join(DEV_DIR, "c_files")
PXD_DIR = os.path.join(DEV_DIR, "pxd_files")














"""This is the *mmgroup* package for computing in the monster group.

In mathematics the *monster group* is the largest sporadic 
finite simple group.  

The most important classes in module ``mmgroup`` are

  ===================  ===========================================
  ``class MM``         Models an element of the monster group.
  ``class MMVector``   Models a vector in a 196884-dimensional
                       representation of the monster group
                       (modulo a small odd number).
  ===================  ===========================================

For documentation see:

https://mmgroup.readthedocs.io/en/latest/
"""

import warnings 


try:
    # Try importing the fast C function
    from mmgroup.mat24 import MAT24_ORDER 
except (ImportError, ModuleNotFoundError):
    # Import a pure python substitute for the mmgroup.mat24 extension
    # if the original extension has not been found. For background, 
    # see section 'Mock up C extension modules' in file
    # docs/source/conf.py.
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24    
    MAT24_ORDER = Mat24.MAT24_ORDER
    del Mat24
    w = "Extension mmgroup.mat24 not found, package not functional!"
    warnings.warn(w, UserWarning)


try:
    from mmgroup.structures.parity import Parity
    from mmgroup.structures.gcode import GCode, GcVector
    from mmgroup.structures.cocode import Cocode
    from mmgroup.structures.ploop import PLoopOne, PLoopOmega
    from mmgroup.structures.ploop import  PLoop, PLoopZ
    from mmgroup.structures.suboctad import Octad, SubOctad
    from mmgroup.structures.autpl import  AutPL
    from mmgroup.structures.xleech2 import XLeech2
except:
    w = "Extension mmgroup.mat24 not found, package not functional!"
    warnings.warn(w, UserWarning)


try:
    import mmgroup.mm
    from mmgroup.mm import INT_BITS
except:
    w = "Extension mmgroup.mm not found, package not functional!"
    warnings.warn(w, UserWarning)
    

try:
    from mmgroup.structures.mm0_group import MM0Group,  MM0
except:
    w = "Class mmgroup.structures.MM0 not found, package not functional!"
    warnings.warn(w, UserWarning)



try:
    import mmgroup.mm_group
    from mmgroup.mm_group import MMGroup, MM
except:
    w = "Class mmgroup.MM not found, package not functional!"
    warnings.warn(w, UserWarning)



import mmgroup.generate_c


try:
    import mmgroup.mm_space
    from mmgroup.mm_space import characteristics
    assert 3 in characteristics()
    from mmgroup.mm_space import MMSpace, MMVector, MMV, order_vector
except:
    w = "Extension mmgroup.mm3 not found, package not functional!"
    warnings.warn(w, UserWarning)

try:
    from mmgroup.structures.xsp2_co1 import Xsp2_Co1
except:
    w = "Module 'mmgroup.structures.xsp2_co1' not found!"
    warnings.warn(w, UserWarning)
    
    


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from math import log
from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.bitfunctions import bit_mat_mul, bit_mat_inverse

from mmgroup.clifford12 import bitmatrix64_t
from mmgroup.clifford12 import bitmatrix64_echelon_h
from mmgroup.clifford12 import bitmatrix64_echelon_l
from mmgroup.clifford12 import bitmatrix64_cap_h
from mmgroup.clifford12 import bitmatrix64_mul
from mmgroup.clifford12 import bitmatrix64_inv
from mmgroup.clifford12 import bitmatrix64_error_pool



 


#####################################################################
# Test function bitmatrix64_t()
#####################################################################


def rand_bit_matrix(rows, cols):
    m = (1 << cols) - 1
    return [randint(0, m) for i in range(rows)]
    
def as_bit_array(m, cols):
    a = np.zeros( (len(m), cols), dtype = np.uint8)
    for i in range(len(m)):
        for j in range(cols):
            a[i,j] = (int(m[i]) >> j) & 1
    return a

            
def create_bitmatrices():
    """yield a bit matrix as a list of integers """
    for rows in range(8):
        for cols in range(8):
            m = rand_bit_matrix(rows, cols)
            yield m, cols
        
        
@pytest.mark.qstate
def test_bitmatrix_t(verbose = 0):
    """Test the transposition of a bit matrix"""
    for ntest, (m, cols) in enumerate(create_bitmatrices()):
        t = bitmatrix64_t(m, cols)
        m1 = as_bit_array(m, cols)
        #print(m1, "shp", m1.shape)
        t1 = as_bit_array(t, len(m))
        if verbose:
            print("Test %d: " % ntest)
            print("Transpose %s bit matrix, m =" % str(m1.shape))
            print(m1, "\nTransposed: %s\n%s" % (str(t1.shape), t1))
        assert m1.T.shape == t1.shape, (m1.shape, t1.shape)
        assert (m1.T == t1).all()


#####################################################################
# Test functions bitmatrix64_echelon_h(), bitmatrix64_echelon_l()
#####################################################################



def create_echelon_matrices():
    """yield a bit matrix as a list of integers """
    for rows in range(8):
        for _ in range(8):
            m = rand_bit_matrix(rows, 64)
            i, j = randint(0,63), randint(0,63)
            yield m, min(i,j), max(i,j)

 
def ref_echelon(m, j0 = 64, n = 64, high = True):
    m = [int(x) for x in m]
    row = 0
    if high:
        assert n <= j0
        range_ = range(j0-1, j0-n-1, -1)
        mask =  ((1 << n) - 1) << (j0 - n)
    else:
        range_ = range(j0, j0+n)
        mask =  ((1 << n) - 1) << (j0)        
    for col in range_:
        col_mask = 1 << col
        for i in range(len(m)-1, row-1, -1):
            if m[i] & col_mask:
                for i1 in range(i-1, -1, -1):
                    if  m[i1] & col_mask: 
                        m[i1] ^= m[i]
                
                m[i], m[row] = m[row], m[i]
                row += 1
                break
    return row, m


@pytest.mark.qstate
def test_bitmatrix_echelon(verbose = 0):
    for ntest, (m, imin, imax) in enumerate(create_echelon_matrices()):
         m1h = np.array(m, dtype = np.uint64, copy = True)
         j0h, nh = imax + 1, imax - imin
         lenh = bitmatrix64_echelon_h(m1h, len(m1h), j0h, nh)
         m1h = list(m1h)
         lenh_ref, m1h_ref =  ref_echelon(m, j0h, nh)
         ok_h = m1h == m1h_ref and lenh == lenh_ref

         m1l = np.array(m, dtype = np.uint64, copy = True)
         j0l, nl = imin, imax - imin
         lenl = bitmatrix64_echelon_l(m1l, len(m1l), j0l, nl)
         m1l = list(m1l)
         lenl_ref, m1l_ref =  ref_echelon(m, j0l, nl, False)
         ok_l = m1l == m1l_ref and lenl == lenl_ref

         ok = ok_h and ok_l
         if verbose or not ok:
              print("Test ", ntest)
              print("m =", [hex(x) for x in m], j0h, nh)
              print("echelon h", [hex(x) for x in m1h], lenh)
              print("expected", [hex(x) for x in m1h_ref], lenh_ref)
              if not ok_h:
                  err = "Error in function bitmatrix64_echelon_h"
                  raise ValueError(err)
              print("m =", [hex(x) for x in m], j0l, nl)
              print("echelon l", [hex(x) for x in m1l], lenl)
              print("expected", [hex(x) for x in m1l_ref], lenl_ref)
              if not ok_l:
                  err = "Error in function bitmatrix64_echelon_l"
                  raise ValueError(err)


#####################################################################
# Test functions bitmatrix64_cap_h()
#####################################################################

def create_cap_matrices():
    """yield a bit matrix as a list of integers """
    test_matrices = [
        ([3,7,11], [3,1], 0,4)
    ]
    for t in test_matrices:
        yield t
    for rows1 in range(5):
        for rows2 in range(5):
            m1 = rand_bit_matrix(rows1, 5)
            m2 = rand_bit_matrix(rows2, 5)
            i, j = randint(0,6), randint(0,6)
            yield m1, m2, min(i,j), max(i,j)
    for rows1 in range(1,65,7):
        for rows2 in range(1,65,7):
            m1 = rand_bit_matrix(rows1, 64)
            m2 = rand_bit_matrix(rows2, 64)
            i, j = randint(0,63), randint(0,63)
            yield m1, m2, min(i,j), max(i,j)




def basis_to_set(b):
    if len(b) == 0:
        return set([0])
    s0 = basis_to_set(b[1:])
    b0 = b[0]
    return s0 | set((x ^ b0 for x in s0))

def as_set(m, mask):
    m = [int(x) & int(mask) for x in m]
    return basis_to_set(m)
        


@pytest.mark.qstate
def test_bitmatrix_cap(verbose = 0):
    for ntest, (m1, m2, imin, imax) in enumerate(create_cap_matrices()):
         if verbose:
             print("Test", ntest+1, "imin =", imin, "imax =", imax )
             print("m1 =", [hex(x) for x in m1])
             print("m2 =", [hex(x) for x in m2])
         m1h = np.array(m1, dtype = np.uint64, copy = True)
         m2h = np.array(m2, dtype = np.uint64, copy = True)
         j0h, nh = imax + 1, imax - imin + 1 
         l1h, l2h = bitmatrix64_cap_h(m1h, m2h, j0h, nh)
         m1h, m2h = list(m1h), list (m2h)
         if verbose:
             print("Non intersecting parts of m1 and m2")
             print("out1 =", [hex(x) for x in m1h[:l1h]])
             print("out2 =", [hex(x) for x in m2h[:l2h]])
             mask = (1 << (imax + 1)) - (1 << imin)
             print("Intersecting parts (mask = %s):" % hex(mask))
             print("out1 =", [hex(x) for x in m1h[l1h:]])
             print("out2 =", [hex(x) for x in m2h[l2h:]])
         if verbose > 1:    
             print("Intermediate results") 
             pool = np.zeros(20, dtype = np.uint64)
             bitmatrix64_error_pool(pool, 20)
             for i in range(20): print(i,  hex(pool[i]))

         assert len(m1) == len(m1h) and len(m2) == len(m2h)
         assert ref_echelon(m1) == ref_echelon(m1h)
         assert ref_echelon(m2) == ref_echelon(m2h)
         mask = (1 << (imax + 1)) - (1 << imin)
         l1, l2 = len(m1), len(m2)
         while l1 and int(m1h[l1-1]) & mask == 0: 
             l1 -= 1
         while l2 and int(m2h[l2-1]) & mask == 0: 
             l2 -= 1
         assert l1-l1h == l2-l2h, (l1, l1h, l2, l2h)
         if len(m1) < 5 and len(m2) < 5:
             set1 = as_set(m1h, mask)
             set2 = as_set(m2h, mask)
             cap  = set1 & set2
             set_cap1 = as_set(m1h[l1h:], mask)
             set_cap2 = as_set(m2h[l2h:], mask)
             assert cap == set_cap1 , (set1, set2, cap, set_cap1)
             assert cap == set_cap2 , (set1, set2, cap, set_cap2)
             if verbose:
                 print("Intersection testd successfully")
            

#####################################################################
# Test functions bitmatrix64_mul() and bitmatrix64_inv()
#####################################################################

# This tests also function bitmatrix64_mask_rows()
# and  bitmatrix64_add_diag().


def create_mul_inv_matrices():
    """yield a bit matrix as a list of integers """
    for i in [1, 3, 5, 7, 17, 24, 31, 32]:
        m2i = None
        for _ in range(1000):
            m2 = rand_bit_matrix(i, i)
            try:
                m2i = bit_mat_inverse(m2)
                break
            except ZeroDivisionError:       
                pass
        for j in [1,2,7,32, 63, 64]:
            m1 = rand_bit_matrix(j, i)
            yield m1, m2, m2i
            m2i = None
             
@pytest.mark.qstate
def test_bitmatrix_mul_inv(verbose = 0):
    for ntest, (m1, m2, m2i) in enumerate(create_mul_inv_matrices()):
         #print(m1, m2, m2i)
         m1a = np.array(m1, dtype = np.uint64, copy = True)
         m2a = np.array(m2, dtype = np.uint64, copy = True)
         m3 =  bit_mat_mul(m1, m2)
         m3a = bitmatrix64_mul(m1a, m2a)
         assert list(m3a) == m3
         if m2i is not None:
            m2ia = bitmatrix64_inv(m2)
            assert list(m2ia) == m2i 


   



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import numpy as np
from random import randint

import pytest

from mmgroup.mm import mm_sub_test_prep_xy
from mmgroup import mat24 as m24

from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.spaces.sparse_mm_space import SparseMmVector
from mmgroup.tests.groups.mgroup_n import MGroupNWord
from mmgroup.mm_space import characteristics


PRIMES = characteristics()




def _as_suboctad(v1, o):
    d = m24.octad_to_gcode(o)
    c = m24.ploop_cap(v1, d)
    return m24.cocode_to_suboctad(c, d)


class prep_xy:
    group = MGroupNWord
    space = SparseMmVector
    def __init__(self, eps, e, f):
        self.f = f & 0x1fff
        self.e = e & 0x1fff
        self.eps = eps = eps & 0xfff
        self.odd = (eps & 0x800) >> 11
        lin = np.zeros(6, dtype = np.uint32) 
        mm_sub_test_prep_xy(f, e, eps, 1, lin)
        self.lin_i = lin[:3]
        self.lin_d = lin[3:6]
        self.sign_XYZ = np.zeros(2048, dtype = np.uint32) 
        mm_sub_test_prep_xy(f, e, eps, 2, self.sign_XYZ)
        self.s_T = np.zeros(759, dtype = np.uint32) 
        mm_sub_test_prep_xy(f, e, eps, 3, self.s_T)


    def inv_op_unit(self, tag, d, j):
        if tag == 'X':
            tag1 = 'X' 
            d1 = d ^ self.lin_d[0] 
            j1 = j
            sign =  (self.sign_XYZ[d] & 1)
            sign ^= (self.lin_i[0] >> j) & 1 
            if self.odd:
                cc = m24.vect_to_cocode(1 << j)
                sign ^=  m24.scalar_prod(d, cc)
        elif tag in 'ZY':
            s = self.odd ^ (tag == 'Y')
            tag1 = 'ZY'[s]
            s += 1
            d1 = d ^ self.lin_d[s]
            j1 = j
            sign =  (self.sign_XYZ[d] >> s) & 1
            sign ^= (self.lin_i[s] >> j) & 1 
        elif tag == 'T':
            tag1 = 'T'
            d1 = d
            te = self.s_T[d] 
            so_exp =  _as_suboctad(self.f, d)
            assert  te & 0x3f == so_exp , (hex(te), hex(so_exp))
            j1 = j ^ (te  & 0x3f)
            sign = m24.suboctad_scalar_prod(j, (te >> 8) & 0x3f)
            sign ^= (te >> 14) & 1
            sign ^= m24.suboctad_weight(j) & self.odd & 1
            assert ((te >> 15) ^ self.odd) & 1 == 0
        else:
            raise ValueError("Illegal tag " + str(tag))
        return sign & 1, tag1, d1, j1

    def inv_op(self, v):
        w = self.space(v.p)
        for value, tag, d, j in v.as_tuples():
            sign, tag, d, j = self.inv_op_unit(tag, d, j) 
            if sign & 1: 
                value = -value % p
            w += value * space(v.p, tag, d, j)
        return w

    def check_v(self, v, verbose = 0):
        grp = self.group
        delta_atom = grp('d', self.eps)
        x_atom = grp('x', self.e)**(-1)
        y_atom = grp('y', self.f)**(-1)
        w_ref = v * delta_atom * x_atom  * y_atom 
        w = self.inv_op(v)
        error = w != w_ref
        if error or verbose:
            eps, e, f = self.eps, self.e, self.f 
            print("vector", v)
            print("operation", "d_%xh * x_%xh * y_%xh" % (eps, e, f))
            print("obtained:", w)
            if error:
                print(v * delta_atom , v, delta_atom)
                print("expected:", w_ref)
                raise ValueError("x-y operation failed")
                print("Error: x-y operation failed!!!")




p = PRIMES[0]
space = SparseMmVector


def as_vector(x):
    if isinstance(x, str):
       data = [(tag, 'r') for tag in x]
       return space(p, data)
    if isinstance(x, tuple):
       return space(p, *x)
    if isinstance(x, list):
       return space(p, x)
    raise TypeError("Bad type for vector of rep")


p = PRIMES[0]
space = SparseMmVector


def prep_xy_testcases():
    testcases = [
       [ [("X", 3, 6)],  0, 0, 0x1171 ],
       [ [("X", 3, 6)],  12, 0, 0 ],
       [ [("X", 3, 6)],  12, 1111, 0 ],
       [ [("X", 3, 6)],  12, 0, 1111],
       [ [("Z", 0, 0)],  0, 0, 0],
       [ [("Z", 0, 0)],  12, 0, 0],
       [ [("Z", 0, 0)],  0, 34, 0],
       [ [("Z", 0, 0)],  0x800, 0, 0],
       [ [("Z", 0, 0)],  0x812, 0, 0],
       [ [("Z", 0, 0)],  0x800, 34, 0],
       [ [("Z", 0, 0)],  0x800, 0, 34],
    ] 
    for v, eps, e, f in testcases:
        yield as_vector(v), prep_xy(eps, e, f) 
    v_tags = "TXZY"
    for v in v_tags:
        for i in range(1000):
            v1 =  as_vector(v)
            eps = randint(0, 0xfff) 
            e = randint(0, 0x1fff)
            f = randint(0, 0x1fff)
            yield v1,  prep_xy(eps, e, f)


@pytest.mark.mm_op
def test_prep_xy(verbose = 0):
    print("Testing preparation of operation x-y...")
    for v, op in prep_xy_testcases():
        op.check_v(v, verbose = verbose)
        if verbose: print("")
    print("passed")
    

    

       



    




from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import warnings
from numbers import Integral


import numpy as np

from mmgroup.generate_c import UserDirective, UserFormat

from mmgroup.dev.mm_basics.mm_basics import INT_BITS
from mmgroup.dev.mm_basics.mm_basics import MM_Const
from mmgroup.structures.abstract_rep_space import mod_inv


MODULI = [7, 31, 127, 255]
N = np.prod(MODULI)
NMAX = N//2
INV = dict(zip(MODULI, [N//m * mod_inv(N//m, m) for m in MODULI]))
# So INV[p] = 1 (mod p),  INV[p] = 0 (mod N/p), p = 7, 31, 127, 255

class MM_CrtCombine:
    TAB_7_31 = np.zeros(256, dtype = np.uint32)
    MASK = 0xffffffff
    for i7 in range(8):
        for i31 in range(32):
            a = (i7 * INV[7] + i31 * INV[31] + NMAX) % N
            TAB_7_31[32 * i7 + i31] = a
    TAB_127 = np.zeros(128, dtype = np.uint32)
    for i127 in range(128):
        a = (i127 * INV[127]) % N - N
        TAB_127[i127] = a & 0xffffffff
    TAB_255 = np.zeros(256, dtype = np.uint32)
    for i255 in range(256):
        a = (i255 * INV[255]) % N - N
        TAB_255[i255] = a & 0xffffffff

    # Chinese remaindering of a7, a31, a127, and a255 works as follows:
    # a = TAB_7_31[32 * a7 + a31] + TAB_127[a127] + TAB_255[a255];
    # a += -((a & 0x80000000) >> 31) & N;
    # a += -((a & 0x80000000) >> 31) & N;
    # a -= NMAX;
    # Now a = a7 (mod7), a = a31 (mod 31), a = a127(mod 127),
    # a = a255(mod 255), -N/2 < a < N/2.
        
    tables = {
        "CRT_TAB_7_31" : TAB_7_31,
        "CRT_TAB_127" : TAB_127,
        "CRT_TAB_255" : TAB_255,
        "CRT_NPROD" : N,
    }

    directives = {}



"""Test C functions for computing the subtype of a Leech lattice vector

In this script we test C functions that compute the subtype of a vector
in the Leech lattice mod 2. These are functions ``gen_leech2_type`` and
``gen_leech2_type2`` in file ``gen_leech.c``.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.

We also count the subtypes of all 0x1000000 vectors in the Leech
lattice modulo 2 and comapare the result against the results in 
:cite:`Iva99`.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__
from multiprocessing import Pool
import time

import numpy as np
import scipy
import scipy.special
from scipy.stats import chisquare

import pytest

from mmgroup import Xsp2_Co1, PLoop, AutPL, Cocode, MM0, MM, XLeech2
from mmgroup.generators import gen_leech2_type
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_type2
from mmgroup.generators import gen_leech2_count_type2
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_type_selftest
from mmgroup.clifford12 import xsp2co1_leech2_count_type2



#####################################################################
# Test computation of subtype Leech lattice vector
#####################################################################

# We test the computation of the subtype of a vetor in the 
# Leech lattice with function gen_leech2_subtype().

# Next we specify certain elements of the Parker loop

# Neutral element of the Parker loop
ZERO = PLoop([])
# The element \Omega of the Parker loop
OMEGA = ~ZERO
# A (preimage of a) dodecad
DODECAD = PLoop([0,4,8, 13,14,15, 17,18,19, 21,22,23])
# A (preimage of an) octad
OCTAD = PLoop([0,1,2,3,4,5,6,7])
# A (preimage of a) complmented octad
HEXADECAD = ~OCTAD


# The following list contains a sample of a vector of each subtype 
# in the Leech lattice mod 2. Such a sample is given as a triple
#   (ploop, cocode, subytpe)
# such that  Xsp2_Co1([("x", ploop), ("d", Cocode(cocode))])
# correspond to a vector of subtype ``subtype`` in the Leech
# lattice modulo 2.
TYPE_DATA = [
    (ZERO, [], 0),
    (OMEGA, [], 0x48),
    (ZERO, [3,4], 0x20),
    (ZERO, [1], 0x21),
    (OCTAD, [], 0x22),
    (OMEGA, [1], 0x31),
    (OCTAD, [7,8], 0x34),
    (OMEGA, [1,2,3], 0x33),
    (DODECAD, [], 0x36),
    (ZERO, [0,1,2,3], 0x40),
    (OCTAD, [0,1], 0x42),
    (ZERO, [1,2,3], 0x43),
    (OCTAD, [8,9], 0x44),
    (DODECAD, [0,1], 0x46),
]


def rand_n_elem():
    """Return random element of the group ``N_x0``

    The group ``N_x0`` has structure ``2^{1+24}.2^{11}.M_24``.

    Multiplying a vector in the Leech lattice mod 2 with an element
    of  ``N_x0`` does not change its subtype.
    """
    return Xsp2_Co1([(x,'r') for x in "dxpy"])


def xs_vector(ploop, cocode):
    """Convert entry of table TYPE_DATA to a Leech lattice vector

    Calling ``xs_vector(ploop, cocode)``, where ``ploop`` and
    ``cocode`` are taken from an entry of the list TYPE_DATA,
    returns the vector in the Leech lattice mod 2 corresponding
    to that entry in Leech lattice encoding.
    """
    return Xsp2_Co1([("x", ploop), ("d", Cocode(cocode))]).as_xsp() 



def display_leech_vector(x):
    """Display a vector in the Leech lattice mod 2

    Here parameter ``x`` is a vector in the Leech lattice mod 2
    in Leech lattice encoding.
    """
    gcode = PLoop(x >> 12)
    bl = gcode.bit_list
    print ("GCode:\n", bl)
    if len(bl) == 16:
        l = [x for x in range(24) if not x in bl]
        pos = bl[0]
    elif len(gcode):
        pos = bl[0]
    else:
        pos = 0
    cocode = Cocode(x) + gcode.theta()
    print ("Cocode:", cocode.syndrome_list(pos))


def alternative_type2(v):
    """Alternative implementation of function ``gen_leech2_type2``

    That function returns 1 if ``v`` is of type 2 and 0 otherwise. 
    """
    a = np.array([0, v], dtype = np.uint64)
    return xsp2co1_leech2_count_type2(a, 2)

def check_leech2_subtype(x, t_expected):
    """Test computation of subtype of vector in Leech lattice mod 2

    Given a vector ``x`` in the Leech lattice mod 2 in **Leech 
    lattice encoding**, the function checkse if the subtype of ``x``
    is correctly computed as the value ``t_expected``.

    Therefore it computes the subtype with function ``gen_leech2_subtype``
    in file ``gen_leech.c`` and checks it against ``subtype``. This 
    function also checks function ``gen_leech2_type2``.   
    """
    t = gen_leech2_subtype(x)
    ok = t == t_expected
    if not ok:
        print("Error: expected Leech type: %s, obtained: %s" % (
            hex(t_expected), hex(t)))
        display_leech_vector(x)
        err = "Error in computing Leech type"
        raise ValueError(err)
    expected_type2 = t_expected if (t_expected >> 4) == 2 else 0
    found_type2 = gen_leech2_type2(x)
    ok = expected_type2 == found_type2
    if not ok:
        print("Error:  x = %s, Leech type: %s" %  
           (hex(x), hex(t_expected)), expected_type2, hex(found_type2))
        err = "Function gen_leech2_subtype2 failed"
        raise ValueError(err)
    alt_found_type2 = alternative_type2(x)
    is_type2 = (t_expected >> 4) == 2
    ok = is_type2 == alt_found_type2
    if not ok:
        print("Error:  x = %s, Leech type: %s" % (
            hex(x),  hex(t_expected)), is_type2, alt_found_type2)
        err = "Function xsp2co1_leech2_count_type2 failed"
        #raise ValueError(err)
    if not is_type2:
       assert not  found_type2 
    assert gen_leech2_type(x) == t >> 4
  

@pytest.mark.gen_xi
def test_leech2_type(verbose = 0):
    """Test computation of subtype of vector in Leech lattice mod 2

    This function takes samples of all subtypes of vectors in
    the Leech lattice mod 2. Any such sample is transformed with
    200 random elements of the group ``N_x0``; this transformation
    does not change the subtype. The we use function 
    ``check_leech2_subtype`` for checking the computation of the
    subtype of all these transformed vectors.
    """
    for ntest, (pl, cocode, vtype) in enumerate(TYPE_DATA):
        x = xs_vector(pl, cocode)
        if verbose:
            print("\nTest %d" %(ntest+1))
            print("Expected type", hex(vtype))
            display_leech_vector(x)
        check_leech2_subtype(x, vtype)
        for i in range(200):
            g =  rand_n_elem()
            x = g.xsp_conjugate(x)
            if verbose:
                display_leech_vector(x)
            check_leech2_subtype(x, vtype)


#*************************************************************************
#** Self test for computing subypes from C file
#************************************************************************/


def binom(n, k):
    """Return binommial coeffient n choose k"""
    return int(scipy.special.binom(n, k) + 0.1)



# The dictionary contains the number TYPE_LENGTHS[t] of vectors
# of subtype t in the Leech lattice modulo 2. This table is 
# obtained from :cite:`Iva99`, Lemmas 4.4.1 and 4.6.1.
TYPE_LENGTHS = {               # Name in :cite:`Iva99`
 0x00: 1,
 0x20: binom(24,2) * 2,        # \Lambda_2^4
 0x21: 24 * 2**11,             # \Lambda_2^3
 0x22: 759 * 2**6,             # \Lambda_2^2
 0x31: 24 * 2**11,             # \Lambda_3^5
 0x33: binom(24,3) * 2**11,    # \Lambda_3^3
 0x34: 759 * 16 * 2**7,        # \Lambda_3^4
 0x36: 2576 * 2**10,           # \Lambda_3^2
 0x40: 2 * 1771,               # \bar{\Lambda}_4^{4a}
 0x42: 759 * 2**6,             # \bar{\Lambda}_4^{6} 
 0x43: binom(24,3) * 2**11,    # \bar{\Lambda}_4^{5}
 0x44: 15 * 759 * 2**7,        # \bar{\Lambda}_4^{4b}
 0x46: 1288 * 2**11,           # \bar{\Lambda}_4^{4c}
 0x48: 1                       # \bar{\Lambda}_4^{8}
}  
assert sum(TYPE_LENGTHS.values()) ==  0x1000000


def one_selftest_leech2(data):
    """Auxiliary function for function ``test_leech2_self``

    This is a wrapper for the C function ``gen_leech2_type_selftest``
    that counts subtypes in an interval of vectors of the Leech
    lattice modulo 2.
    """
    start, n = data
    a = np.zeros(0x50, dtype = np.uint32)
    result =  gen_leech2_type_selftest(start, n, a)
    return result, a


def gen_selftest_inputs(n):
    """Auxiliary function for function ``test_leech2_self``

    Generates inputs for function ``one_selftest_leech2`` 
    for multiprocessing.
    """
    assert 0x1000000 % n == 0
    q = 0x1000000 // n
    for  i in range(n):
        yield i*q, q


@pytest.mark.gen_xi
@pytest.mark.slow
def test_leech2_self(verbose = 0):
    """Test number of entries of all subtypes

    We count the subtypes of all 0x1000000 vector in the Leech
    lattice modulo 2, an we compile a dictionary that contains 
    the number of such vectors for each subtype. The we compare
    that dictionary against the dictionary TYPE_LENGTHS.

    We use the C function ``gen_leech2_type_selftest`` for 
    accelerating that computation, and we also use 
    multiprocessing.    
    """
    NPROCESSES = 4
    with Pool(processes = NPROCESSES) as pool:
        results = pool.map(one_selftest_leech2, 
                   gen_selftest_inputs(NPROCESSES))
    pool.join()
    result = sum(x[0] for x in results)
    a = np.zeros(0x50, dtype = np.uint32)
    for x in results:
         a += x[1]
    d = {}
    for i, n in enumerate(a):
        if n:
            d[i] = n
    if verbose:
        print("Type-4 vectors: %d\nVector types:" % result)
        for t, n in d.items():
            print(" %2x: %7d" % (t, n))
    assert sum( TYPE_LENGTHS.values() ) == 2**24
    for t, n in d.items():
        assert n == TYPE_LENGTHS[t]
    N4 = sum(n for i, n in d.items() if i & 0xf0 == 0x40)
    assert result == N4






#####################################################################
# Test counting short vectors in the Leech lattice mod 2
#####################################################################


@pytest.mark.xsp2co1
def test_leech2_count(fast = 1, verbose = 0):
    f = xsp2co1_leech2_count_type2 if fast else gen_leech2_count_type2
    a_type = np.uint64 if fast else np.uint32
    if verbose:
        print("Testing gen_leech2_type2() ... ")
    a = np.array([0]+[1 << i for i in range(24)], dtype = a_type)
    t_start = time.time()
    result = f(a, 25)
    t = time.time() - t_start
    assert result == 98280, result
    if verbose:
        print("passed, %.3f ms" % (1000*t))

#####################################################################
# Test method subtype of class Xsp2Co1
#####################################################################


def subtype_testdata():
    yield XLeech2(~PLoop())
    yield XLeech2(Cocode([0,1,2,3]))
    yield XLeech2(~PLoop(list(range(8))))
    yield XLeech2(~PLoop(list(range(8))), Cocode([8,9]))
    for i in range(50):
        yield XLeech2('r', 4)


@pytest.mark.xsp2co1
def test_subtype(verbose = 0):
    OMEGA = XLeech2(~PLoop())
    if verbose:
         print("OMEGA = ", OMEGA)
    types = set()
    for v in subtype_testdata():
         g = MM('c', v)
         v2 =  Xsp2_Co1(g)
         v2_subtype = v2.subtype
         v2ref = (OMEGA * g)
         v2ref_subtype = v2ref.subtype
         if verbose:
              print("v = ", v)
              print("g = ", g)
              print("v2 = ", v2, ", subtype =", v2_subtype)
              print("v2ref = ", v2ref, ", subtype =", v2ref_subtype)
         assert v2_subtype== v2ref_subtype, (v2_subtype, v2ref_subtype)
         types.add(v2_subtype)
    assert len(types) == 6













import sys
import os
import numpy as np
from numbers import Integral
from random import randint
from random import choice


import pytest

from mmgroup.tests.groups.mgroup_n import MGroupN
from mmgroup.structures.mm0_group import MM0Group, MM0, StdMM0Group



####################################################################
# Conversion between MM0 (i.e. the Monster group) and MGroupN 
# (i.e. an implementation of the subgroup N_0 of the monster).
####################################################################

ref_group = MGroupN()       # Implementation of subgroup N_0 of the monster
group = MM0Group()          # Implementation of the monster group
assert group is StdMM0Group # Check that class MMGroup is a singleton.




####################################################################
# Test cases
####################################################################

def rand_word(tags, length):   
    return [(choice(tags), 'r') for i in range(length)]


def make_testcases():
    testdata = [
        ([("t", 2), ("x",0)], []),
        ([("d", 2), ("d",2)], []),
        ([("p", 220249159)], []),
        ([("y", 0x1000),('p', 234579)], [("y", 0x1800)]),
        ([("y", 0x1000),('t',2)], [("y", 0x1800)]),
        ([("x", 0x800), ('t',1)],  [("y", 0x5a2), ("y", 0x1da2)]),  
    ]
    for d in testdata:
        yield d
    for i in range(5):
        for g1 in "dpxyt":
            for g2 in "dpxyt":
                yield [(g1,'r')], [(g2,'r')]
    for i in range(150):
        g1 = rand_word("dpxyt", 10)
        g2 = rand_word("dpxyt", 10)
        yield g1, g2


####################################################################
# Test group multiplication
####################################################################

def ll_mul_test_case(g1, g2, verbose = 0):
    g1, g2 = ref_group(g1), ref_group(g2)
    g1 = MM0(g1)
    g2 = MM0(g2)
    g1ref = ref_group(g1) 
    g2ref = ref_group(g2) 
    g12ref = g1ref * g2ref
    g12 = g1 * g2
    if verbose:
        print("g1 =", g1)
        print("g2 =", g2)
        print("g1 * g2 =", ref_group(g12))
        print("ref     =", g12ref)
    assert MM0(g12ref) == g12, (g1, g2, g12ref, g12)
    assert ref_group(g12) == g12ref, (g1, g2, ref_group(g12), g12ref)


@pytest.mark.auto_group
def test_ll_mul(verbose = 0):
    print("testing low-level multiplication in group N")
    for g1, g2 in make_testcases():
        ll_mul_test_case(g1, g2, verbose)
    print("passed")



####################################################################
# Test group division 
####################################################################

def ll_div_test_case(g1, g2, verbose = 0):
    g1 = MM0(g1)
    g2 = MM0(g2)
    g1ref = ref_group(g1) 
    g2ref = ref_group(g2) 
    g12ref = g1ref / g2ref
    g12 = g1 / g2
    assert MM0(g12ref) == g12, (g12ref, g12)
    assert ref_group(g12) == g12ref


@pytest.mark.auto_group
def test_ll_div(verbose = 0):
    print("testing low-level multiplication in group N")
    for g1, g2 in make_testcases():
        ll_div_test_case(g1, g2, verbose)
    print("passed")





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


from numbers import Integral
import numpy as np
from random import randint

import scipy
from scipy import stats

import pytest

from mmgroup.mm_space import MMSpace, MMVector
from mmgroup.mm_space import characteristics
from mmgroup.mm import mm_aux_index_sparse_to_extern
from mmgroup.mm import mm_aux_index_extern_to_sparse
from mmgroup.mm import mm_aux_mmv_extract_sparse

from mmgroup.tests.spaces.spaces import MMTestSpace
from mmgroup.tests.spaces.sparse_mm_space import SparseMmSpace
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.spaces.sparse_mm_space import SparseMmVector





#########################################################################
### Chi square test
#########################################################################

def chisquare_crosscor(p, b1, b2 = None, d = 1, m = 0):
    """chi-square test over cross correlation
    
    Let b1 and b2 be arrays of intgers (mod p) of the same length.
    the we put a[i] = b1[i] + b2[i + d] (mod p), with idex i + d 
    wrapping around. b2 may be 0, indicating an array of zeros.

    We return the chisquare test probability assumint that b1 and b2
    are equidistributed.

    In case m > 0 the entries a[i] are reduced modulo m. 
    """
    d %= len(b1)
    if b2 is None:
        a =  np.zeros(len(b1), dtype = np.uint32)
    else:
        assert len(b1) == len(b2)
        a =  np.array(np.concatenate((b2[d:], b2[:d])), dtype = np.uint32)
    a = (a + b1) % p
    f_exp = None
    if 0 < m < p:
        a %= m
        f = np.fromiter((x % m for x in range(p)), dtype = np.uint32)
        f_exp = np.bincount(f, minlength=m)
        f_exp = f_exp * len(a) / p
        p = m
    bins = np.bincount(a, minlength=p)
    chisq, prob = scipy.stats.chisquare(bins, f_exp)
    return prob

def chisquare_crosscor_ok(p, b1, b2 = 0, d = 1, m = 0):
    for i in range(4):
        prob = chisquare_crosscor(p, b1, b2, d, m)
        if 0.01 < prob < 0.99: return True, prob
    return False, prob


#########################################################################
### Test random vectors and manipulation of them
#########################################################################

def do_test_sparse_rep(v):
    p = v.p
    space = MMVector
    sparse_space = SparseMmVector
    vsp = sparse_space(p, v)
    assert (v == space(p, vsp))
    assert (v == space(p, "S", v.as_sparse()))
    v2 = space(p, 'V', v.as_bytes())
    assert (v == v2)

def do_test_rep_conversion(v):
    p = v.p
    space = MMVector
    sparse_space = SparseMmVector
    ranges = [
        (0, 5, 1),
        (0, 50, 10),
        (randint(0,23), 852, randint(13, 17)),
        (randint(852,852+200), 196884, randint(390, 409)),
    ]
    for  start, stop, step in ranges:
        data_sparse =  v.as_bytes()[start:stop:step] 
        data = v[start:stop:step]
        assert len(data_sparse) == len(data), (len(data_sparse), len(data))
        assert (data_sparse == data).all(), (data_sparse, data)
        a = np.zeros(len(data), dtype = np.uint32)
        for i, index in enumerate(range(start,stop,step)):
            a[i] =  mm_aux_index_extern_to_sparse(index)             
            i_ext =  mm_aux_index_sparse_to_extern(a[i]) 
            assert i_ext == index, (hex(index), hex(a[i]), hex(i_ext))
        mm_aux_mmv_extract_sparse(p, v.data, a, len(a))
        assert (a & 0xff == data).all(), (a & 0xff, data)



def check_direct_symmetric(v):
    for tag in "ABC":
        m = v[tag]
        assert (m == m.T).all()
        if tag in "BC":
            for i in range(24):
                assert m[i,i] == 0 
        assert v[tag, 3, 3] == m[3, 3]
        for k in range(4):
            i0, i1 = randint(0, 23), randint(0, 23)
            assert  v[tag, i0, i1] == m[i0, i1]
   
def do_test_random_io(p, verbose = False, to_sparse = 1):
    name_slices = [
        ("D", "D", 1),
        ("T", "T", 1),
        ("As",  slice(24,24+276), 1),
        ("Bs",  slice(24+276,24+2*276), 1),
        ("Cs",  slice(24+2*276,24+3*276), 1),
        ("ABCs",slice(0,24+3*276), 1), 
        ("X", "X", 1),
        ("Y", "Y", 1),
        ("Z", "Z", 1),
        ("A", "A", 0),
        ("B", "B", 0),
        ("C", "C", 0),
    ]
    chisqu_ok = {}
    for (name, _, do_chisqu_test) in name_slices: 
        chisqu_ok[name] = not do_chisqu_test
    for i in range(4):
        space = MMTestSpace(p)
        v = space(0)
        assert np.count_nonzero(v.data[:v.ops.MMV_INTS]) == 0
        v.check()
        v = space('R')
        v.check()
        check_direct_symmetric(v)
        if to_sparse and i == 0:
            do_test_sparse_rep(v)
        do_test_rep_conversion(v)
        
        for name, slice_, do_chisqu_test in name_slices:
            v1 = v[slice_]
            if name  in "DABCTXYZ":
                xx = v[name]
                assert (v1 % p == xx % p).all(), (p, name, slice_)
            if do_chisqu_test:
                v1 = v1.reshape(-1)
                prob1 =  chisquare_crosscor(p, v1,  m = 23)
                prob2 =  chisquare_crosscor(p, v1, v1, 7,  m = 11)
                if verbose:
                    print("p=%3d, %-8s: %.8f, %.8f" % (p,name,prob1,prob2))
            pr = [prob1, prob2]
            if 0.01 < min(pr) <= max(pr) < 0.99: 
                chisqu_ok[name] = 1 
    errors = ", ".join([ x for x in chisqu_ok if not chisqu_ok[x] ])
    if len(errors):
        print("\nChisqu test failed for p=%d and %s !!\n" % (p,errors))
        raise ValueError("Chisquare test failed for p=%d " % p)

@pytest.mark.mm_op
def test_all_random_io(verbose = 0):
    print("Chisquare test for randomization of an mm vector")
    for i in range(2):
        to_sparse = 0 if i else 1
        for p in characteristics():
            do_test_random_io(p, verbose, to_sparse)
    print("Test passed")



#########################################################################
### Test manipulation of symmetric tags A, B, C
#########################################################################


def do_test_sym_io(p, tag, verbose = 1):
    def array_data(shape,  data):
        if len(shape) == 0:
            return data 
        b1 = np.fromiter(data, dtype = np.int32).reshape(shape) 
        return b1         
    a_data = [
        [(slice(0,3), slice(0,3)), (3,3), 
                  (i*j+1 for i in range(3) for j in range(3))],
        [(slice(2,5), 17), (3,), 
                  (i for i in range(3))],
        [(slice(2,4), slice(2,4)), (2,2), 
                   (i*j+1 for i in range(2) for j in range(2))],
        [(slice(2,24), slice(2,24)), (22,22), 
                 (i*j+1 for i in range(22) for j in range(22))],
    ]               
    if verbose:
        print("p = %d, tag = %s" % (p,tag))
    space = MMTestSpace(p)
    b = np.zeros((24,24), dtype=np.int32)
    v = space(0)
    for slices, shape, data in a_data:
        b1 = array_data(shape, data)
        v[tag, slices[0], slices[1]] = b1
        b[slices[0], slices[1]] = b1 % p
        b1T = b1.T if len(shape) == 2 else b1 
        b[slices[1], slices[0]] = b1T % p
        
    if tag in "BC":
        for i in range(24): b[i,i] = 0
    vd = v[tag]
    assert (vd == vd.T).all()
    assert (b % p == vd).all(), (tag, b - vd, b, vd)

    for (sl0, sl1), shape, _ in a_data:
        eq = v[tag, sl0, sl1] == vd[sl0, sl1]
        if len(shape): eq = eq.all()
        assert eq

    abc = v[:852]
    v1 = space(0)
    v1[:852] = abc
    assert v == v1, (v, v1)


def do_test_sym_io_bad(p, tag, verbose = 1):
    def array_data(shape, value):
        if len(shape) == 0:
            return value
        return np.zeros(shape, dtype = np.int32) + value
    a_data = [
        [(slice(1,23), slice(2,24)), (22,22), 1],
        [(slice(0,11), slice(9,19)), (11,10), 2],
        [(1,23), (), 4],
        [(slice(7,11), 23), (4,),  8],
    ]
    if verbose:
         print("p = %d, tag = %s, bad data" % (p,tag))
    space = MMTestSpace(p)
    v = space(0)
    w = np.zeros( (24, 24), dtype = np.int32)
    for slices, shape, value in a_data:
        new_data = array_data(shape, value)
        v[tag, slices[0], slices[1]] = new_data
        w[slices[0], slices[1]] = new_data
        v.check()
        try:        
            new_data_t = new_data.T
        except:
            assert isinstance(new_data, Integral) or len(new_data.shape) < 2
            new_data_t = new_data
        w[slices[1], slices[0]] = new_data_t
        w = w % p
        if tag in "BC":
            for i in range(24): w[i,i] = 0
        assert (v[tag] == w).all(), (tag, v[tag] - w, v[tag], w)
           



@pytest.mark.mm_op
def test_sym_all(verbose = 0):
    print("Test io operations for tags ABC")
    for tag in "ABC":
        for p in characteristics():
            do_test_sym_io(p, tag, verbose = verbose)
            do_test_sym_io_bad(p, tag, verbose = verbose)
    print("Test passed")


#########################################################################
### Test manipulation of tags T, X, Y, Z
#########################################################################


def do_test_large_io(p, tag, verbose = 0):
    def array_data(shape,  data):
        if len(shape) == 0:
            return data 
        b1 = np.fromiter(data, dtype = np.int32).reshape(shape) 
        return b1         
    t_d = [
        (slice(0, 3, 1), slice(0, 5, 1)),
        (slice(18, 759, 20), slice(2,64, 3)),
        (slice(19,622, 70), 53),
        (52, slice(0,64,9)),
        (313,55), 
    ]
    x_d = [
        (slice(18, 2048, 79), slice(2,24, 7)),
        (slice(19, 1622, 70), 23),
        (1552, slice(0,24,9)),
        (1313,15), 
    ]
    lengths, a_data = ((759,64), t_d) if tag =='T' else ((2048,24), x_d)
    v = MMTestSpace(p)('R')
    for slices in a_data:
        prod, shape = 1, ()
        for l, s in zip(lengths, slices):
            if isinstance(s, slice):
                size_ = len(range(*s.indices(l)))
                prod *= size_
                shape = shape + (size_,)
        data = np.array(p * np.random.rand(prod), dtype = np.uint32)
        if len(shape) == 0:
            data = data[0]
        else:
            data = data.reshape(shape)
        v[tag, slices[0], slices[1]] = data
        eq = v[tag, slices[0], slices[1]] == data
        if len(shape): eq = eq.all()
        assert eq, (tag, v[tag, slices[0], slices[1]], data)
    v.check()

@pytest.mark.mm_op
def test_large_all(verbose = 0):
    print("Test io operations for tags TXYZ")
    for tag in "TXYZ":
        for p in characteristics():
            do_test_large_io(p, tag, verbose = verbose)
    print("Test passed")







"""Auxiliary functions for writing tables into C program"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import re
import os
import warnings
import ast


def pxi_defs(pxifile_name):
    """Return dictionary containing definitions in a .pxi file.

    .pxi files used in the Cython environment contain lines of shape:

    DEF <name> = <literal>

    The function returns a dictionary with entries <name> : <literal>
    """
    f = open(pxifile_name, "rt")
    m_def = re.compile(r"\s*DEF\s+(\w+)\s*=\s*(.+)")
    d = {}
    for l in f:
        try:
            name, value = m_def.match(l).groups()
            d[name] = ast.literal_eval(value.strip())
        except:
           pass
    f.close()
    return d





m_pxd_line = re.compile(r"\s+(\w+)\s+(\w+)\((.+)\)")
m_pxd_arglist = re.compile(r"\s*(\w+)(\s*(\*)?\s*)(\w+)")

def _parse_pxd_line(l):
    m = m_pxd_line.match(l) 
    if m:
         type, function, args = m.groups()
         arglist = args.split(",")
         args = []
         for arg in arglist:
             m_arg = m_pxd_arglist.match(arg)
             if not m_arg: return None
             g = m_arg.groups()
             args.append( (g[0],) + g[2:])
         return (type, function, args) 


m_pxd_enable = re.compile(r"\s*#\s+PYX(.+)")

def _parse_pxd_enable(l):
    m = m_pxd_enable.match(l) 
    if m:
        return True
    else:
        return False



def pxd_to_pyx(pxd_file, module = None, translate = None, select = True, nogil = False): 
    """Extract Cython wrappers from prototypes in a .pxd file

    A .pxd file contains prototypes of external C functions and it
    may be included into a .pyx file in order to create a Cython 
    module that uses these C functions.

    This function returns a string containing python wrappers of the
    C functions in the pxd file. This string may be used as a part of 
    an automatically generated .pyx file. Here the C functions in the
    .pxd file must be sufficiently simple as indicated below.

    The returned string starts with two lines::
 
        cimport cython 
        cimport <module>
    
    Here <module> is given by the parameter ``module``, By default,
    ``module`` is constructed from the name ``pxd_file`` of the .
    pxd file.

    The .pxd file is parsed for lines of shape::

       <return_type> <function>(<type1> <arg1>,  <type2> <arg2> , ...)

    Every such line is converted to a string that codes function,
    which is a Cython wrapper of that function of shape::

       @cython.boundscheck(False)  # Deactivate bounds checking
       @cython.wraparound(False)   # Deactivate negative indexing.
       def <translated_function>(<arg1>, <arg2> , ...):
           cdef <type1> <arg1> = <arg1>_v_
           cdef <type2> <arg2> = <arg2>_v
           cdef <return_type>  _ret
           ret_ = <module>.<function>(<arg1>_v_,  <arg2>_v_ , ...)
           return ret_

    ``<translated_function>`` is the string computed as 
    ``translate(<function>)``, if the argument ``translate`` is given.
    Otherwise is ``<translated_function>`` is equal to ``<function>``.
    
    ``<return_type>`` and ``<type1>, <type2>, ...`` must be valid types 
    known to Cython and C. The types  ``<type1>, <type2>`` used for 
    arguments may  also be pointers. Then pointers are converted to
    memory views e.g::

        uint32_t <function>(uint8_t *a)

    is converted to::

       @cython.boundscheck(False)  # Deactivate bounds checking
       @cython.wraparound(False)   # Deactivate negative indexing.
       def <translated_function1>(a):
           cdef uint8_t[::1] a_v_ = a
           cdef uint32_t ret_
           ret_ = <function>(&a_v_[0])
           return ret_

    The ``<return_type>`` may be void, but not be a pointer. Other 
    types  are not allowed.

    If parameter ``select`` is True then we convert a function only
    if a line containing the string "``# PYX``" (possibly with
    leading blanks) precedes the declaration of a function.
    By default, ``select`` is True. In the code generator you may 
    use the ``EXPORT`` directive with options ``px`` to enter such
    a line into the source file.

    If ``nogil`` is True, a C function is called with as follows::

       with nogil:
           ret_ = <function>(&a_v_[0])

    Then the C function  ``<function>`` must be e.g. declared as 
    follows::
 
        cdef extern from "<file>.h" nogil:
            int <function> (int a, int *b)
      
    This feature releases the GIL (Global Interpreter Lock) when
    calling the function.
    """
    if not module:
        module = os.path.splitext(os.path.basename(pxd_file))[0]
    s = "cimport cython\n"
    s += "cimport %s\n\n" % module
    if nogil:
        s += "cimport cython\n"
    enable = not select
    with open(pxd_file, "rt") as input_file:
        for l in input_file:
            data = _parse_pxd_line(l)
            if not data or not enable:
                if select:
                     enable = _parse_pxd_enable(l)
                continue
            return_type, function, args = data
            t_function = translate(function) if translate else function
            s += "\n"
            s += "@cython.wraparound(False)\n"
            s += "@cython.boundscheck(False)\n"
            s += "def {f}({args}):\n".format( 
              f = t_function, args = ", ".join([a[2] for a in args]))
            c_args = []
 
            has_returnvalue = return_type != "void"
            for argtype, is_ptr, name in args:
                memview = "[::1]"  if is_ptr else ""
                s += "    cdef {argtype}{memview} {name}_v_ = {name}\n".format(
                    argtype=argtype, memview=memview, name=name)
                if is_ptr:
                    c_args.append("&{name}_v_[0]".format(name=name))
                else:
                    c_args.append("{name}_v_".format(name=name))
            assign = ""
            if has_returnvalue:
                s += "    cdef {type} ret_\n".format(type=return_type)
                assign = "ret_ = "
            if nogil:
                s += "    with nogil:\n    "
            s += "    {assign}{m}.{f}({args})\n".format(
                assign = assign, m = module, f = function, 
                args = ", ".join(c_args)
                ) 
            if has_returnvalue:
                s += "    return ret_\n"
            enable = not select
    return s


def pxd_to_function_list(pxd_file): 
    """Extract function names from prototypes in a .pxd file

    Returns a list of function names found in a .pxd file.
    This list contains exactly the names of the functions that
    would be wrapped by calling ``pxd_to_pyx(pxd_file)``.
    """
    l_out = []
    with open(pxd_file, "rt") as input_file:
        for l_in in input_file:
            data = _parse_pxd_line(l_in)
            if  data:
                _0, function, _1 = data
                l_out.append(function)
    return l_out






import os
import sys
import time
import numpy as np
import re


import_pending = True


def import_all():
    global MMV15, import_pending
    global MM0,  MMV, MMVector, MMSpace
    global get_order_vector, order_vector_from_dict
    from mmgroup.structures.mm0_group import MM0
    from mmgroup.mm_space import MMV, MMVector, MMSpace
    from mmgroup.dev.mm_reduce.order_vector import get_order_vector
    from mmgroup.dev.mm_reduce.order_vector import order_vector_from_dict
    MMV15 = MMV(15)
    import_pending = False
  

NAMES = [
   "TAG_DATA", 
   "S_G71", "S_V71", "S_GA", "DIAG_VA", "S_G94", "S_V94",
   "TAGS_Y", "TAGS_X", "TAG_SIGN"
]


def order_vector_from_py_data():
    if import_pending:
        import_all()
    ov, o_tag, data = get_order_vector()
    data[ "TAG_DATA"] = o_tag
    return ov, o_tag, data


def order_vector_from_C_data():
    if import_pending:
        import_all()
    data = {}
    LEN = 128
    from mmgroup.mm_reduce import  mm_order_load_tag_data
    from mmgroup.mm_reduce import  mm_order_load_vector
    for i, name in enumerate(NAMES):
        a = np.zeros(LEN, dtype = np.uint32)
        length = mm_order_load_tag_data(i, a, LEN)
        data[name] = a[:length]
    o_tag = data["TAG_DATA"]
    ov = MMV15(0)
    mm_order_load_vector(ov.data)
    return ov, o_tag, data



def compress_order_vector_data(ov):
    t = []
    v = ov.data
    def compress24(start, length):
        for i in range(start, start + 2*length, 2):
             t.append(int(v[i]) & 0xffffffff)
             t.append(int(v[i]) >> 32)
             t.append(int(v[i+1]) & 0xffffffff)
    def compress64(start, length):
        for i in range(start, start + 4*length):
             t.append(int(v[i]) & 0xffffffff)
             t.append(int(v[i]) >> 32)
    OFS_ABC = 0
    OFS_T = 72*2
    OFS_XYZ = OFS_T + 759*4
    compress24(OFS_ABC, 72)
    compress64(OFS_T, 759)
    compress24(OFS_XYZ, 3*2048)
    return np.array(t, dtype = np.uint32)


class OrderVectorTable:
    directives = {}
    tables = None

    def compute_data(self):
        """Compute data for table classes

        We cannot to this ab initio; because the required inputs 
        are not available if sys.arg contains tan argument ``mockup``.
        """
        if not self.__class__.tables is None:
            return
        if import_pending:
            import_all()
        ov, tag, data =  order_vector_from_py_data()  
        a_ov =  compress_order_vector_data(ov)
        tag_data = []
        tag_indices = [0]
        for i, name in enumerate(NAMES):
            a0 = list(data[name])
            tag_data += a0
            tag_indices.append(tag_indices[-1] + len(a0))
        tag_data = np.array(tag_data, dtype = np.uint32)
        tag_indices = np.array(tag_indices, dtype = np.uint16)
        self.__class__.tables = {
            "ORDER_VECTOR_DATA": a_ov,
            "ORDER_VECTOR_TAG_DATA": tag_data,
            "ORDER_VECTOR_TAG_INDICES": tag_indices,
            "ORDER_VECTOR_NUM_TAG_DATA": len(NAMES),
            }
    
    def __init__(self, *args):
        self.compute_data()


class Mockup_OrderVectorTable:
    a_ov =  np.array([0], dtype = np.uint32)
    tag_data = np.array([0], dtype = np.uint32)
    tag_indices = np.array([0,0], dtype = np.uint16)
    tables = {
        "ORDER_VECTOR_DATA": a_ov,
        "ORDER_VECTOR_TAG_DATA": tag_data,
        "ORDER_VECTOR_TAG_INDICES": tag_indices,
        "ORDER_VECTOR_NUM_TAG_DATA": 1
    }
    directives = {}

    def __init__(self, *args):
        pass



def check_order_vector_data():
    if import_pending:
        import_all()
    from mmgroup.mm_reduce import  mm_order_compare_vector
    from random import randint

    ov, o_tag, data = order_vector_from_C_data()
    ov_ref, o_tag_ref, _ = order_vector_from_dict(data)
    ov, ov_ref = ov.data, ov_ref.data
    if not np.array_equal(ov, ov_ref):
        for i, w in enumerate(ov):
            w_ref = ov_ref[i]
            if w != w_ref:
                print("Order vector error at index %d" % i)
                print("Found: %s, expected: %s" % (hex(w), hex(w_ref)))
                raise ValueError("Error in order vector")
    assert np.array_equal(o_tag, o_tag_ref)
    if (mm_order_compare_vector(ov_ref)):
       err = "Comparing with order vector failed" 
       raise ValueError(err) 
    positions = [(0,0), (72*2, 63), (72*2 + 759*4 - 1, 35)]
    for i in range(30, len(ov) - 1, 537):
        bmax = 63 if 72*2 < i < 72*2 + 759*4 else 31
        positions.append( (i, randint(0,bmax)) ) 
    for i, sh in positions:
         a = np.copy(ov_ref)
         a[i] = int(a[i]) ^ (1 << sh)
         result = mm_order_compare_vector(a)  
         if (result != 1):
             err = "Comparing of bad vector with order vector failed" 
             raise ValueError(err) 
              





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import re
import warnings
import copy
from random import randint, randrange, sample

from copy import deepcopy
from numbers import Integral

from mmgroup.structures.parse_atoms import AtomDict 
from mmgroup.structures.parse_atoms import eval_atom_expression, ihex 


######################################################################
# Utility functions for scalars
######################################################################




def mod_inv(a, p):
    """Return x with x * a == 1 (mod p)  for integers a, p.

    p >= 1 must hold.
    Raise ZerodivsionError if inversion is not possible
    """
    assert p >= 1
    a1, a2, x1, x2 = p, a % p, 0, 1
    # Invert a modulo p with the extended Euclidean algorithm
    while a2 > 0:
        # assert x1 * a % p == a1 % p and  x2 * a % p == a2 % p
        q, r = divmod(a1, a2)
        a1, a2, x1, x2 = a2, r, x2, x1 - q * x2
    if a1 != 1:
        raise ZeroDivisionError("Cannot invert %d mod %d" % (a, p))
    # assert x1 * a % p == 1
    return x1 % p


def mod_exp(a, e, p):
    """Return a**e (mod p) for integers  a, e, p  with p >= 1.

    Raise ZeroDivsionError if this is not possible
    """
    assert p >= 1
    return pow(a, e, p) if e >= 0 else pow(mod_inv(a, p), -e, p) 


def mod_pwr2(e, p):
    """Return 2**e (mod p) for integers e, p with p >= 1.

    Raise ZeroDivsionError if this is not possible. This is
    faster than  mod_exp(2, e, p)  for negative e.
    """
    assert p >= 1
    if e >= 0:
        return pow(2, e, p)
    elif p & 1:
        return pow((p + 1) >> 1, -e, p)
    else:
        raise ZeroDivisionError("Cannot invert 2 mod %d" % p)
 

def mod_rand_invertible(p):
    """Return a random invertible element modulo p for  p > 1"""
    assert p > 1
    while True:  
        a = randint(1, p - 1)
        # Check with the Euclidean algorithm if a is invertible
        a2, a1 = p, a
        while a1 > 0:
            a2, a1 = a1, a2 % a1
        # Now a is invertible mod p iff a1 == 1
        if a2 == 1:
            return a


def mod_rand_unit(p):
    """Return a random unit 1 or -1 modulo p for  p > 1"""
    assert p > 1
    return 1 + (p - 2) * randint(0, 1)
    



######################################################################
# class AbstractRepVector
######################################################################



class AbstractRepVector(object):
    """Model an vector in an abstract representation vector space 

    Here a vector space is an instance of class AbstractRepSpace
    """
    __slots__ = "space"
    def __init__(self, space, *args, **kwds):
        self.space = space

    def __eq__(self, other):    
        return( isinstance(other, AbstractRepVector) 
            and  self.space == other.space
            and  self.space.equal_vectors(self, other)
        )

    def __ne__(self, other): 
        return not self.__eq__(other)

    def copy(self):
        """Return deep copy of vector 'self'"""
        return self.space.copy_vector(self)

    def __imul__(self, other):
        if  isinstance(other, Integral):
            return self.space.imul_scalar(self, other) 
        else:
            return self.space.imul_group_word(self, other)

    def __mul__(self, other):
        return self.copy().__imul__(other)

    def __rmul__(self, other):
        return self.space.imul_scalar(self.copy(), other) 

    def __itruediv__(self, other):
        if  isinstance(other, Integral):
            p = self.p
            try:
                return self.__imul__(mod_inv(other, p)) 
            except ZeroDivisionError:
                if p != 0:
                    raise
                res = self
                if other < 0:
                    res = self.__imul__(-1) 
                    other = -other
                bl = p.bit_length()
                if p != (1 << bl) >> 1:
                    raise
                return res.__ilshift__(other)
                    
        else:
            other = self.space.group(other)
            return self.space.imul_group_word(self, other**(-1))

    def __truediv__(self, other):
        return self.copy().__itruediv__(other)


    def __ilshift__(self, other):
        try: 
            factor = mod_pwr2(other, self.p)
            return self.space.imul_scalar(self, factor) 
        except ZeroDivisionError:
            if p != 0:
                 raise

    def __lshift__(self, other):
        return self.copy().__ilshift__(other)

    def __irshift__(self, other):
        return self.__ilshift__(-other)

    def __rshift__(self, other):
        return self.copy().__ilshift__(-other)

    def __neg__(self):
        return self.space.imul_scalar(self.copy(), -1) 

    def __pos__(self):
        return self  

    def __iadd__(self, other):
        if other in self.space:
            return self.space.iadd(self, other)
        raise TypeError("Vectors must be in the same vector space")

    def __add__(self, other):
        return self.copy().__iadd__(other)

    def __isub__(self, other):
        return self.__iadd__(self, other.__neg__())

    def __sub__(self, other):
        return self.copy().__iadd__(other.__neg__())

    def __getitem__(self, index):
        return self.space.vector_get_item(self, index) 

    def __setitem__(self, index, value):
        return self.space.vector_set_item(self, index, value) 
        
    def reduce(self):
        return self.space.reduce(self)

    def str(self):
        try:
            return self.space.str_vector(self)
        except NotImplementedError:
            return super(AbstractRepVector, str)()

    __repr__ = str

######################################################################
# class AbstractRepSpace
######################################################################


class AbstractRepSpace(object):
    group = None                    # group operating on the space
    vector_type = AbstractRepVector # type of an vector in the space 
    atom_parser = {}                # see method parse()
    FRAME = re.compile(r"^(\w*)$")  # see method parse()

    def __init__(self, p, *args, **kwds):
        """Creating instances is only possible for concrete spaces

         
        """
        self.p = p


    ### The following methods must be overwritten ####################

    def zero(self):
        """Return the zero vector of the vector space"""
        raise NotImplementedError("No unit vector in abstract space")

    def unit(self,  *args):
        """Return a unit vector of the vector space

        Constructing a unit vector without any arguments should
        return the zero vector.
        """
        raise NotImplementedError("No unit vector in abstract space")

    def iadd(self, v1, v2):
        """Return sum v1 + v2 of vectors v1 and v2.

        Vector v1  may be destroyed, but not v2.

        This method is called for vectors v1, v2 of the space
        'self'  only.
        """
        raise NotImplementedError("No vector addition in abstract space")


    def imul_scalar(self, v1, a):
        """Return product a*v1 of scalar a and vector v1.

        v1 may be destroyed.

        This method is called for elements v1 of the space
        'self' only.

        Here scalars are integers. In case self.p > 0 scalars
        are integers modulo p.

        Here we a bit sloppy, not claiming that p is prime.
        So 'self' may also be a free module over the integers
        modulo p for composite p.
        """
        raise NotImplementedError("No scalar multiplication in abstract space")


    def imul_group_word(self, v1, g):
        """Return product v1*g of vector v1 and group word g.

        v1 may be destroyed.

        This method is called for elements v1 of the space
        'self' and for elements g of the group 'self.group' only.
        """
        raise NotImplementedError("No group multiplication in abstract space")




    def vector_get_item(self, v1, index):
        """Return v1[index] for a vector v1 in the vector space"""
        raise NotImplementedError("Cannot get component in abstract space")

    def vector_set_item(self, v1, index, value) :
        """Put v1[index] = value for a vector v1 in the vector space"""
        err = "Item assignment not supported in space of type '%s'"
        raise NotImplementedError(err % type(self))

    def equal_vectors(self, v1, v2):
        """Return True iff vectors v1 and v2 are equal 

        This method is called for elements v1 and v2 of the space
        'self' only.
        """
        raise NotImplementedError("Cannot test equality in abstract space")


    ### The following methods should be overwritten ###################

    def copy_vector(self, v1):
        """Return deep copy of group element v1"""
        v2 = deepcopy(v1)
        v2.space = v1.space
        return v2


    def iadd_tuple(self, v1, t):
        """Compute v += self.unit(*t)

        In a concrete space, this may be optimized considerably.
        """
        raise NotImplementedError("Not supported")
        return self.iadd(v1, self.unit(*t))


    def reduce(self, v):
        """Reduce the vector v which is an element of the space

        We assume that the representation of a vector
        is not always given in a unique form that we call
        the reduced form. 

        This method tries to achieve this goal. Vectors
        may be reduced by any operator, even by '==',
        without notice.
        """
        return v

    def str_vector(self, v):
        """Convert vector v to a string

        """
        raise NotImplementedError


    def ishift_right(self, v1, shift):
        """Return product v1 / 2**shift .

        v1 may be destroyed.

        This method is called for elements v1 of the space
        'self' in case ``v1.p == 0`` only.

        The function should return ``V1 * 2**(-shift)`` is this is 
        defined.
        """
        raise NotImplementedError("Division by two not supported")


    ### The following methods need not be overwritten #################


    def parse(self, s):
        """Convert the string 's' to a vector
        
        Here 's' may ba any valid python expression. Then certain 
        identifiers are recognized e.g. as unit vectors of the 
        space, and operations given by the string 's' are performed 
        on the values of these identifiers. The function tries to 
        convert the result  of the evaluated python expression to 
        a  vector using the standard conversion functions 
        in the __call__  method of this group.

        The identifiers to be recognized are given by the dictionary 
        'self.atom_parser'. This dictionary is empty in the abastract 
        space but it may be set to a specific directory in a 
        specific space.

        A suitable atom parser is AtomDict(self.unit), with AtomDict 
        taken from module parse_atoms.py. This parser maps certain 
        identifiers to tuples and it calls method self.unit() with
        the entries of that tuple as arguments.

        If the string 's' matches the match object self.FRAME, then
        the string self.FRAME.match(s)[1] is parsed instead of 's'.
        This feature may be used to unpack the string 's', e.g. 
        for reversing the effect of a formatting method for 
        vectors. See the standard module re.py for match objects.

        Of course the user may overwrite this method with his own
        parser.
        """
        m = self.FRAME.match(s)
        if m: s = m[1]
        return eval_atom_expression(s, self.atom_parser)

    ### The following methods should not be overwritten ###############

    def set_preimage(self, space, morphism):
        """Set a natural 'morphism' from 'space' to 'self'.

        Here 'morphism' must be a natural homomorphism that maps 
        elements of the space 'space' to elements of the space 
        'self'. 'space' must be an instance of class
        AbstractRepSpace.

        If v is an element of 'space' then self(v) returns
        morphism(v), which is an element of the space self.
        So we may call class 'self', which represents a vector
        space, to map elements of other spaces to the space 'self'. 
        """
        raise NotImplementedError("Not supported")
        self.preimages[space] = morphism


    def __contains__(self, other):
        """Return True iff other is an element of the space"""
        try:
            if not isinstance(other, AbstractRepVector): 
                 return False
            return other.space == self
        except:
            return False

           

    def __call__(self,  *args):
        """Construct a vector of the rep space 'self'.

        The returned vector is the sum of all arguments 
        given to the function. Here each argument v  is 
        converted to a vector as follows:

        - An element v of this space is not changed.

        - If v is an element of another space V and there is a 
          natural hmomorphism phi from V to this space (set by
          method set_preimage) then v is converted to phi(v).

        - A tuple v is converted to self.unit(*tuple).

        - The integer 0 then is converted to the zero vector.

        - If v is a string, we convert v to self.parse(v),
          and perform one of the actions listed above.

        If none of these condition is met, we raise TypeError.
        """
        raise NotImplementedError
        result = self.zero()
        for v in args:
            if isinstance(v, str):
                v = self.parse(v)
            if isinstance(v, tuple):
                self.iadd_tuple(result, v)
            elif isinstance(v, AbstractRepVector):
                if v.space != self:
                    try:
                        morphism = self.preimages[v.space]
                    except KeyError:
                        err = "Cannot map vector to given vector space"
                        raise TypeError(err)
                    v = morphism(v)
                self.iadd(result, v)
            elif v == 0:
                pass
            else:
                err = "Cannot convert '%s' object to a rep vector"
                raise TypeError(err % str(type(v)))
        return self.reduce(result)


















